"use strict";
/**
 * @author Sean sean.snow@live.com
 * @date 2015/12/25
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.BitSet = void 0;
const Long_1 = require("../lang/Long");
const StringBuilder_1 = require("../lang/StringBuilder");
const Arrays_1 = require("./Arrays");
const ADDRESS_BITS_PER_WORD = 5;
const BITS_PER_WORD = 1 << ADDRESS_BITS_PER_WORD;
const WORD_MASK = 0xffffffffffffffff;
class BitSet {
    constructor(nbits) {
        this.words = [];
        this.sizeIsSticky = false;
        this.wordsInUse = 0;
        if (nbits < 0) {
            throw Error(`nbits < 0: ${nbits}`);
        }
        this._initWords(nbits);
        this.sizeIsSticky = true;
    }
    set(bitIndex) {
        if (bitIndex < 0) {
            throw Error(`bitIndex < 0: ${bitIndex}`);
        }
        let wordIndex = this._wordIndex(bitIndex);
        this._expandTo(wordIndex);
        this.words[wordIndex] |= (1 << bitIndex);
    }
    get(bitIndex) {
        if (bitIndex < 0) {
            throw Error(`bitIndex < 0: ${bitIndex}`);
        }
        let wordIndex = this._wordIndex(bitIndex);
        return (wordIndex < this.wordsInUse) && ((this.words[wordIndex] & (1 << bitIndex)) != 0);
    }
    length() {
        if (this.wordsInUse === 0) {
            return 0;
        }
        return BITS_PER_WORD * (this.wordsInUse - 1) +
            (BITS_PER_WORD - Long_1.Long.numberOfLeadingZeros(this.words[this.wordsInUse - 1]));
    }
    size() {
        return this.words.length * BITS_PER_WORD;
    }
    cardinality() {
        let sum = 0;
        for (let i = 0; i < this.wordsInUse; i++) {
            sum += Long_1.Long.bitCount(this.words[i]);
        }
        return sum;
    }
    nextSetBit(fromIndex) {
        if (fromIndex < 0) {
            throw Error(`Index Out Of Bounds Exception. fromIndex < 0: ${fromIndex}`);
        }
        let u = this._wordIndex(fromIndex);
        if (u >= this.wordsInUse) {
            return -1;
        }
        let word = this.words[u] & (WORD_MASK << fromIndex);
        while (true) {
            if (word != 0) {
                return (u * BITS_PER_WORD) + Long_1.Long.numberOfLeadingZeros(word);
            }
            if (++u === this.wordsInUse) {
                return -1;
            }
            word = this.words[u];
        }
    }
    nextClearBit(fromIndex) {
        if (fromIndex < 0) {
            throw Error(`Index Out Of Bounds Exception. fromIndex < 0: ${fromIndex}`);
        }
        let u = this._wordIndex(fromIndex);
        if (u >= this.wordsInUse)
            return fromIndex;
        let word = ~this.words[u] & (WORD_MASK << fromIndex);
        while (true) {
            if (word != 0)
                return (u * BITS_PER_WORD) + Long_1.Long.numberOfTrailingZeros(word);
            if (++u == this.wordsInUse)
                return this.wordsInUse * BITS_PER_WORD;
            word = ~this.words[u];
        }
    }
    _initWords(nbits) {
        this.words = new Array(this._wordIndex(nbits - 1) + 1);
    }
    _wordIndex(bitIndex) {
        return bitIndex >> ADDRESS_BITS_PER_WORD;
    }
    _expandTo(wordIndex) {
        let wordsRequired = wordIndex + 1;
        if (this.wordsInUse < wordsRequired) {
            this._ensureCapacity(wordsRequired);
            this.wordsInUse = wordsRequired;
        }
    }
    _ensureCapacity(wordsRequired) {
        if (this.words.length < wordsRequired) {
            // Allocate larger of doubled size or required size
            let request = Math.max(2 * this.words.length, wordsRequired);
            this.words = Arrays_1.Arrays.copyOf(this.words, request);
            this.sizeIsSticky = false;
        }
    }
    toString() {
        let numBits = (this.wordsInUse > 128) ? this.cardinality() : this.wordsInUse * BITS_PER_WORD;
        let b = new StringBuilder_1.default(6 * numBits + 2);
        b.append('{');
        let i = this.nextSetBit(0);
        if (i !== -1) {
            b.append(i);
            while (true) {
                if (++i < 0)
                    break;
                if ((i = this.nextSetBit(i)) < 0)
                    break;
                let endOfRun = this.nextClearBit(i);
                do {
                    b.append(", ").append(i);
                } while (++i != endOfRun);
            }
        }
        b.append('}');
        return b.toString();
    }
}
exports.BitSet = BitSet;
exports.default = BitSet;
